/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */


package org.ow2.petals.databinding.jaxb.serialization;

import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.Unmarshaller;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.databinding.jaxb.DataBindingException;
import org.ow2.petals.databinding.jaxb.service.Message;
import org.ow2.petals.databinding.jaxb.service.Operation;
import org.ow2.petals.databinding.jaxb.tools.DataBindingConstants;
import org.ow2.petals.databinding.jaxb.tools.Util;

/**
 * 
 * @author Mohammed EL JAI
 *
 */
public class ServiceUnmarshaller {


    /** The service class loader **/
    private ClassLoader serviceClassLoader;
    
    /** Class loader to load the generated JAXB classes**/
    private ClassLoader jaxbClassLoader;




    /**
     * 
     * @param serviceOperation
     * @param source
     * @return
     * @throws DataBindingException
     */
    public Object[] unmarshallMethodParam(Operation serviceOperation, Source source) throws DataBindingException {

        Object[] methodParams = null;

        try {



            Message inputMessage = serviceOperation.getInputMessage();
            Unmarshaller unmarshaller = serviceOperation.getServiceOfOperation().getRegistry().getUnmarshaller(
                    inputMessage.getMessageName().getNamespaceURI());

            // Create message object that represents the input message
            Object jaxbMessageObject = null;
            try {
                if (source instanceof StreamSource)
                {
                    jaxbMessageObject = unmarshaller.unmarshal(((StreamSource) source).getInputStream());
                }
                else
                {
                    jaxbMessageObject = unmarshaller.unmarshal(source);
                }
            }
            catch(javax.xml.bind.UnmarshalException e) {
                File request = new File("request-temp.xml");
                Result result = new StreamResult(request);
                TransformerFactory tsfF = TransformerFactory.newInstance();
                Transformer trsf = tsfF.newTransformer();
                trsf.transform(source, result);
                jaxbMessageObject = unmarshaller.unmarshal(request);
                request.delete();
            }
            if (jaxbMessageObject instanceof JAXBElement) {
                JAXBElement<?> jaxbElement = (JAXBElement<?>) jaxbMessageObject;
                jaxbMessageObject = jaxbElement.getValue();
            }


            Map<String,Class<?>> messageParts = inputMessage.getMessageParts();
            methodParams = new Object[inputMessage.getMessageParts().size()];

            Class<?> messageJaxbClass = jaxbMessageObject.getClass();

            Class<?>[] methodParamTypes = serviceOperation.getParameterTypes();

            Type[] genericParamTypes = serviceOperation.getGenericParameterTypes();

            int paramIndex = 0;
            for (String messgePartName : messageParts.keySet()) {

                Field field = messageJaxbClass.getDeclaredField(messgePartName);
                String getterMethodName = null;


                if (!field.getType().equals(boolean.class) && !field.getType().equals(java.lang.Boolean.class)) {
                    getterMethodName = "get" + Util.upperFirstCase(messgePartName);
                }
                // In the case where the field type is a boolean, the getter method name start with "is" prefix
                else {
                    getterMethodName = "is" + Util.upperFirstCase(messgePartName);
                }

                Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName, new Class[]{});

                // Current parameter is a simple type or primitive type
                if (methodParamTypes[paramIndex].isPrimitive() || Util.isSimpleType(methodParamTypes[paramIndex])){
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        methodParams[paramIndex] = getValueMethod.invoke(getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        methodParams[paramIndex] = getterMethod.invoke(jaxbMessageObject);
                }

                // Current parameter is a Calendar type
                else if (methodParamTypes[paramIndex].getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getValueMethod.invoke(
                                getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getterMethod.invoke(jaxbMessageObject);
                        
                    java.util.Calendar date = null;

                    if (jaxbDate != null) {
                        date = jaxbDate.toGregorianCalendar();
                    }
                    methodParams[paramIndex] = date;
                }

                // Current parameter is a Calendar type
                else if (methodParamTypes[paramIndex].getName().equals(DataBindingConstants.DATE_CLASS_NAME)) {
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getValueMethod.invoke(
                                getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getterMethod.invoke(jaxbMessageObject);
                    
                    java.util.Date date = null;
                    if (jaxbDate != null) {
                        date = jaxbDate.toGregorianCalendar().getTime();

                    }
                    methodParams[paramIndex] = date;
                }

                // Current parameter is a list
                else if (java.util.Collection.class.isAssignableFrom(methodParamTypes[paramIndex]) || 
                        methodParamTypes[paramIndex].isArray()) {

                    Object jaxbArrayObject = getterMethod.invoke(jaxbMessageObject, new Object[]{});
                    if (jaxbArrayObject.getClass().isArray()) {
                        methodParams[paramIndex] = jaxbArrayObject;
                    }
                    else {
                        Collection<?> jaxbList;
                        if (!java.util.Collection.class.isAssignableFrom(jaxbArrayObject.getClass())) { 
                            Method getListMethod = jaxbArrayObject.getClass().getDeclaredMethods()[0];
                            jaxbList = (Collection<?>) getListMethod.invoke(jaxbArrayObject, new Object[0]);
                        }
                        else {
                            jaxbList = (Collection<?>) jaxbArrayObject;
                        }
                        if (!methodParamTypes[paramIndex].isArray())
                            methodParams[paramIndex] = createOriginalList(jaxbList,genericParamTypes[paramIndex]);
                        else {
                            Collection<?> originalList = createOriginalList(jaxbList,genericParamTypes[paramIndex]);
                            Class<?> elementType = methodParamTypes[paramIndex].getComponentType();
                            try {
                                Object[] tab = (Object[]) java.lang.reflect.Array.newInstance(elementType, originalList.size());
                                methodParams[paramIndex] = originalList.toArray(tab);
                            }
                            catch(java.lang.ClassCastException e) {
                                methodParams[paramIndex] = Util.ListToprimitiveArray((List<?>) originalList, elementType);
                            }

                        }
                    }
                }

                // current parameter is a Bean
                else {
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        methodParams[paramIndex] = createOriginalBean(getValueMethod.invoke(getterMethod.invoke(jaxbMessageObject)),
                                methodParamTypes[paramIndex]);
                    }
                    else
                        methodParams[paramIndex] = createOriginalBean(getterMethod.invoke(jaxbMessageObject, new Object[]{}),
                                methodParamTypes[paramIndex]);
                }
                paramIndex ++;
            }
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }

        return methodParams;
    }




    /**
     * 
     * @param serviceOperation
     * @param source
     * @return
     * @throws DataBindingException
     */
    public Object unmarshallMethodResult(Operation serviceOperation, Source source) throws DataBindingException {

        Object[] methodResult = null;

        try {
            Message outputMessage = serviceOperation.getOutputMessage();
            Unmarshaller unmarshaller = serviceOperation.getServiceOfOperation().getRegistry().getUnmarshaller(
                    outputMessage.getMessageName().getNamespaceURI());

            // Create message object that represents the input message
            Object jaxbMessageObject = null;
            try {
                if (source instanceof StreamSource)
                {
                    jaxbMessageObject = unmarshaller.unmarshal(((StreamSource) source).getInputStream());
                }
                else
                {
                    jaxbMessageObject = unmarshaller.unmarshal(source);
                }
            }
            catch(javax.xml.bind.UnmarshalException e) {
                File request = new File("request-temp.xml");
                Result result = new StreamResult(request);
                TransformerFactory tsfF = TransformerFactory.newInstance();
                Transformer trsf = tsfF.newTransformer();
                trsf.transform(source, result);
                jaxbMessageObject = unmarshaller.unmarshal(request);
                request.delete();
            }
            if (jaxbMessageObject instanceof JAXBElement) {
                JAXBElement<?> jaxbElement = (JAXBElement<?>) jaxbMessageObject;
                jaxbMessageObject = jaxbElement.getValue();
            }


            Map<String,Class<?>> messageParts = outputMessage.getMessageParts();
            methodResult = new Object[outputMessage.getMessageParts().size()];

            Class<?> messageJaxbClass = jaxbMessageObject.getClass();

            Class<?> methodReturnType = serviceOperation.getReturnType();

            Type genericReturnType = serviceOperation.getGenericReturnType();

            int paramIndex = 0;
            for (String messgePartName : messageParts.keySet()) {

                String jaxbFieldName = messgePartName;
                if (jaxbFieldName.equals("return")) jaxbFieldName = "_return";
                Field field = messageJaxbClass.getDeclaredField(jaxbFieldName);
                String getterMethodName = null;
                if (!field.getType().equals(boolean.class) && !field.getType().equals(java.lang.Boolean.class)) {
                    getterMethodName = "get" + Util.upperFirstCase(messgePartName);
                }
                else {
                    getterMethodName = "is" + Util.upperFirstCase(messgePartName);
                }

                Method getterMethod = messageJaxbClass.getDeclaredMethod(getterMethodName, new Class[]{});

                // Current parameter is a primitive type
                if (methodReturnType.isPrimitive() || Util.isSimpleType(methodReturnType)){
                    
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        methodResult[paramIndex] = getValueMethod.invoke(getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        methodResult[paramIndex] = getterMethod.invoke(jaxbMessageObject);
                    
                }

                // Current parameter is a Calendar type
                else if (methodReturnType.getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {
                    
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getValueMethod.invoke(
                                getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getterMethod.invoke(jaxbMessageObject);
                    
                    java.util.Calendar date = null;

                    if (jaxbDate != null) {
                        date = jaxbDate.toGregorianCalendar();
                    }
                    methodResult[paramIndex] = date;
                }

                // Current parameter is a Calendar type
                else if (methodReturnType.getName().equals(DataBindingConstants.DATE_CLASS_NAME)) {
                    
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getValueMethod.invoke(
                                getterMethod.invoke(jaxbMessageObject));
                    }
                    else
                        jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) getterMethod.invoke(jaxbMessageObject);
                    
                    java.util.Date date = null;
                    if (jaxbDate != null) {
                        date = jaxbDate.toGregorianCalendar().getTime();

                    }
                    methodResult[paramIndex] = date;
                }

                // Current parameter is a list
                else if (java.util.Collection.class.isAssignableFrom(methodReturnType) ||
                        methodReturnType.isArray()) {
                    Object jaxbArrayObject = getterMethod.invoke(jaxbMessageObject, new Object[]{});
                    if (jaxbArrayObject.getClass().isArray()) {
                        methodResult[paramIndex] = jaxbArrayObject;
                    }
                    else {
                        Collection<?> jaxbList;
                        if (!java.util.Collection.class.isAssignableFrom(jaxbArrayObject.getClass())) { 
                            Method getListMethod = jaxbArrayObject.getClass().getDeclaredMethods()[0];
                            jaxbList = (Collection<?>) getListMethod.invoke(jaxbArrayObject, new Object[0]);
                        }
                        else {
                            jaxbList = (Collection<?>) jaxbArrayObject;
                        }
                        if (!methodReturnType.isArray())
                            methodResult[paramIndex] = createOriginalList(jaxbList,genericReturnType);
                        else {
                            Collection<?> originalList = createOriginalList(jaxbList,genericReturnType);
                            Class<?> elementType = methodReturnType.getComponentType();
                            try {
                                Object[] tab = (Object[]) java.lang.reflect.Array.newInstance(elementType, originalList.size());
                                methodResult[paramIndex] = originalList.toArray(tab);
                            }
                            catch(java.lang.ClassCastException e) {
                                methodResult[paramIndex] = Util.ListToprimitiveArray((List<?>) originalList, elementType);
                            }

                        }
                    }
                }

                // Current parameter is a Bean
                else {
                    
                    String getValueMethodName = "getValue";
                    Method getValueMethod = null;
                    if (field.getType().equals(JAXBElement.class)) {
                        getValueMethod = field.getType().getDeclaredMethod(getValueMethodName);
                        methodResult[paramIndex] = createOriginalBean(getValueMethod.invoke(getterMethod.invoke(jaxbMessageObject)),
                                methodReturnType);
                    }
                    else
                        methodResult[paramIndex] = createOriginalBean(getterMethod.invoke(jaxbMessageObject, new Object[]{}),
                                methodReturnType);
                    
                }
                paramIndex ++;
            }
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }

        return methodResult[0];

    }


    /**
     * 
     * @param serviceOperation
     * @param source
     * @return
     * @throws DataBindingException
     */
    public Object unmarshallException(Operation serviceOperation, Source source) throws DataBindingException {

        try {
            
            JAXBContext jaxbContext = JAXBContext.newInstance(
                    com.sun.tools.xjc.api.XJC.getDefaultPackageName(serviceOperation.getOperationFaults().get(0)
                            .getMessageName().getNamespaceURI())
                    , jaxbClassLoader);
            Unmarshaller exceptionUnmarshaller = jaxbContext.createUnmarshaller();
            
            // Create message object that represents the fault message
            Object jaxbExceptionObject = null;
            try {
                if (source instanceof StreamSource)
                    jaxbExceptionObject = exceptionUnmarshaller.unmarshal(((StreamSource) source).getInputStream());
                else
                    jaxbExceptionObject = exceptionUnmarshaller.unmarshal(source);
            }
            catch(javax.xml.bind.UnmarshalException e) {
                File request = new File("request-temp.xml");
                Result result = new StreamResult(request);
                TransformerFactory tsfF = TransformerFactory.newInstance();
                Transformer trsf = tsfF.newTransformer();
                trsf.transform(source, result);
                jaxbExceptionObject = exceptionUnmarshaller.unmarshal(request);
                request.delete();
            }
            
            if (jaxbExceptionObject instanceof JAXBElement) {
                JAXBElement<?> jaxbElement = (JAXBElement<?>) jaxbExceptionObject;
                jaxbExceptionObject = jaxbElement.getValue();
            }

            Class<?> exceptionClass = null;
            for (Class<?> clazz : serviceOperation.getExceptionType()) {
                if (clazz.getName().equalsIgnoreCase(jaxbExceptionObject.getClass().getName()))
                    exceptionClass = clazz;
            }
            return (Exception) createOriginalBean(jaxbExceptionObject, exceptionClass);
        }

        catch(Exception e) {
            throw new DataBindingException(" Error occurs while unmarshalling the operation fault ",e);
        }

    }


    /**
     * 
     * @param jaxbBeanInstance
     * @return
     * @throws DataBindingException 
     */
    private Object createOriginalBean(Object jaxbBeanInstance, Class<?> originalBeanClazz) throws DataBindingException {

        if (jaxbBeanInstance == null) {
            return null;
        }

        Object originalBeanInstance = null;
        try {

            Class<?> originalBeanClass = null;
            try {
                originalBeanClass = serviceClassLoader.loadClass(jaxbBeanInstance.getClass().getName());
            }
            catch (Exception e) {
                // The case when the namespace and the package name are different
                if (originalBeanClazz != null)
                    originalBeanClass = serviceClassLoader.loadClass(originalBeanClazz.getName());
                else 
                    throw new DataBindingException(e);
            }

            originalBeanInstance = originalBeanClass.newInstance();
            Field[] originalBeanFields = Util.getFieldFormBeanClassAndSuperClass(originalBeanClass);


            for (Field originalBeanField : originalBeanFields) {

                String fieldName =  originalBeanField.getName();

                if (fieldName.equals("serialVersionUID")) 
                    continue;

                Field[] jaxbBeanFields = Util.getFieldFormBeanClassAndSuperClass(jaxbBeanInstance.getClass());
                Field jaxbField = Util.getJaxbField(fieldName, jaxbBeanFields);
                if (jaxbField == null)
                    continue;

                String getterMethodName = "get" + Util.upperFirstCase(jaxbField.getName());
                Method getterMethod  = null;
                try {
                    getterMethod = jaxbBeanInstance.getClass().getMethod(getterMethodName, new Class[]{});
                }
                catch (java.lang.NoSuchMethodException e) {
                    // The field type is a boolean, it's getter method starts with "is"
                    getterMethodName = "is" + Util.upperFirstCase(jaxbField.getName());
                    getterMethod = jaxbBeanInstance.getClass().getMethod(getterMethodName, new Class[]{});
                }

                String setterMethodName = "set" + Util.upperFirstCase(fieldName);
                Method setterMethod = Util.getMethodFomBeanClassOrSuperClass(setterMethodName, 
                        originalBeanClass, originalBeanField.getType());


                String getValueMethodName = "getValue";	
                Object jaxbElement = getterMethod.invoke(jaxbBeanInstance, new Object[]{});	

                if (jaxbElement == null) {
                    //setterMethod.invoke(originalBeanInstance, new Object[]{null});
                    continue;
                }



                Method getValueMethod = null;
                if (jaxbElement instanceof javax.xml.bind.JAXBElement) 
                    getValueMethod = jaxbElement.getClass().getDeclaredMethod(getValueMethodName, new Class[]{});


                // Current parameter is a primitive type
                if (originalBeanField.getType().isPrimitive() || Util.isSimpleType(originalBeanField.getType())){
                    if (getValueMethod != null)
                        setterMethod.invoke(originalBeanInstance, getValueMethod.invoke(jaxbElement, new Object[]{}));
                    else 
                        setterMethod.invoke(originalBeanInstance, jaxbElement);
                }


                // Current parameter is a Calendar type
                else if (originalBeanField.getType().getName().equals(DataBindingConstants.CALENDAR_CLASS_NAME)) {
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) 
                    getValueMethod.invoke(jaxbElement, new Object[]{});
                    java.util.Calendar date = jaxbDate.toGregorianCalendar();
                    setterMethod.invoke(originalBeanInstance, date);
                }

                // Current parameter is a Calendar type
                else if (originalBeanField.getType().getName().equals(DataBindingConstants.DATE_CLASS_NAME)) {
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) jaxbElement;
                    java.util.Date	date = jaxbDate.toGregorianCalendar().getTime();
                    setterMethod.invoke(originalBeanInstance, date);
                }

                // Current parameter is a list
                else if (java.util.Collection.class.isAssignableFrom(originalBeanField.getType()) ||
                        originalBeanField.getType().isArray()) {
                    Object jaxbArrayObject;
                    if (getValueMethod != null) {
                        jaxbArrayObject = getValueMethod.invoke(jaxbElement, new Object[]{});
                        if (jaxbArrayObject == null)
                            continue;
                    }
                    else 
                        jaxbArrayObject = jaxbElement;
                    if (jaxbArrayObject.getClass().isArray()) {
                        setterMethod.invoke(originalBeanInstance, new Object[]{jaxbArrayObject});
                    }
                    else {
                        Collection<?> jaxbList;
                        if (!java.util.Collection.class.isAssignableFrom(jaxbArrayObject.getClass())) { 
                            Method getListMethod = jaxbArrayObject.getClass().getDeclaredMethods()[0];
                            jaxbList = (Collection<?>) getListMethod.invoke(jaxbArrayObject, new Object[0]);
                        }
                        else {
                            jaxbList = (Collection<?>) jaxbArrayObject;
                        }       
                        if (!originalBeanField.getType().isArray())
                            setterMethod.invoke(originalBeanInstance,  createOriginalList(jaxbList,originalBeanField.getGenericType()));
                        else {
                            Collection<?> originalList = createOriginalList(jaxbList,originalBeanField.getGenericType());
                            Class<?> elementType = originalBeanField.getType().getComponentType();
                            try {
                                Object[] tab = (Object[]) java.lang.reflect.Array.newInstance(elementType, originalList.size());
                                setterMethod.invoke(originalBeanInstance, new Object[]{originalList.toArray(tab)});
                            }
                            catch(java.lang.ClassCastException e) {
                                setterMethod.invoke(originalBeanInstance, 
                                        new Object[]{Util.ListToprimitiveArray((List<?>) originalList, elementType)});
                            }
                        }
                    }
                }

                // Current parameter is a Bean
                else {
                    if (getValueMethodName != null)
                        setterMethod.invoke(originalBeanInstance, createOriginalBean(getValueMethod.invoke(jaxbElement, new Object[]{}),
                                originalBeanField.getType()));
                    else 
                        setterMethod.invoke(originalBeanInstance, createOriginalBean(jaxbElement,
                                originalBeanField.getType()));
                }

            }
        } catch (Exception e) {
            throw new DataBindingException(e);
        } 
        return originalBeanInstance;
    }


    /**
     * Create collection from JAXB list 
     * @param jaxbList The JAXB list 
     * @return The created collection
     * @throws DataBindingException 
     * @throws IllegalAccessException 
     * @throws InstantiationException 
     */
    @SuppressWarnings("unchecked")
    private Collection<?> createOriginalList(Collection<?> jaxbList, Type collectionClass) throws DataBindingException {

        if (jaxbList == null) return null;

        Collection<Object> originalList = null;

        try {

            Class<?> genericTypeClass = null;
            Type genericType = null;

            // Create the original list instance
            // if (collectionClass == null) 
            //   originalList = java.util.ArrayList.class.newInstance();
            //else {
            if (collectionClass instanceof Class<?>) {
                if (((Class<?>) collectionClass).isArray()) {
                    originalList = java.util.ArrayList.class.newInstance();
                    genericTypeClass = ((Class<?>) collectionClass).getComponentType();
                }
                else 
                    if (! ((Class<?>)collectionClass).isInterface()){
                        originalList = (Collection<Object> ) ((Class<?>)collectionClass).newInstance();
                    }else{
                        originalList = java.util.ArrayList.class.newInstance();
                    }
            }
            else if (collectionClass instanceof ParameterizedType) {

                ParameterizedType pt = (ParameterizedType) collectionClass;
                Class<?> originalListClass = (Class<?>) pt.getRawType();
                if (! originalListClass.isInterface()){
                    originalList = (Collection<Object> ) originalListClass.newInstance();
                }else{
                    originalList = java.util.ArrayList.class.newInstance();
                }
                genericType = pt.getActualTypeArguments()[0];

                if (genericType instanceof Class<?>) {
                    genericTypeClass = (Class<?>) genericType;
                }

                else if (genericType instanceof ParameterizedType) {
                    genericTypeClass = (Class<?>) ((ParameterizedType) genericType).getRawType();
                }

                else 
                    throw new DataBindingException(" Unknown type :  " + genericType.getClass().getName());
            }
            else 
                throw new DataBindingException(" Unknown type :  " + collectionClass.getClass().getName());
            //}
            // Browse the JAXB list elements 
            for (Object listContent : jaxbList) {

                if (listContent == null)
                    continue;

                if (listContent.getClass().isPrimitive() || Util.isSimpleType(listContent.getClass())){
                    originalList.addAll(jaxbList);
                    break;
                }

                // Current parameter is a Calendar type
                else if (listContent.getClass().getName().equals(DataBindingConstants.XMLCALENDAR_CLASS_NAME)) {
                    javax.xml.datatype.XMLGregorianCalendar jaxbDate = (javax.xml.datatype.XMLGregorianCalendar) listContent;
                    java.util.Calendar date = null;
                    if (jaxbDate != null) {
                        date = jaxbDate.toGregorianCalendar();
                    }
                    if (genericTypeClass != null && genericTypeClass.getName().equals(DataBindingConstants.DATE_CLASS_NAME))
                        originalList.add(date.getTime());
                    else 
                        originalList.add(date);
                }


                // The list element is a collection type
                else if (genericTypeClass != null && java.util.Collection.class.isAssignableFrom(genericTypeClass)) {
                    originalList.add(createOriginalList((Collection<?>) listContent,genericType));
                }

                // The list element is a bean
                else {
                    originalList.add(createOriginalBean(listContent, genericTypeClass));
                }
            }
        }
        catch(Exception e) {
            throw new DataBindingException(e);
        }

        return originalList;
    }

    public ClassLoader getServiceClassLoader() {
        return serviceClassLoader;
    }


    public void setServiceClassLoader(ClassLoader serviceClassLoader) {
        this.serviceClassLoader = serviceClassLoader;
    }
    

    public ClassLoader getJaxbClassLoader() {
        return jaxbClassLoader;
    }

    
    public void setJaxbClassLoader(ClassLoader jaxbClassLoader) {
        this.jaxbClassLoader = jaxbClassLoader;
    }

}
