/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */


package org.ow2.petals.databinding.jaxb.tools;

import java.util.List;

import java.io.File;

import com.sun.tools.xjc.Driver;

/**
 * 
 * @author Mohammed EL JAI
 * 
 * This class offers some utilities to generate and compile the JAXB classes 
 */

public class JaxbClassGenerator {
	
	
	/**
	 * Generate the JAXB classes for the WSDL given in parameter 
	 * @param wsdlUrl : The WSDL URL
	 * @param rootPath : The base directory of the generation
	 * @param genDirName : The name of the directory where the sources will be generated 
	 */
	public static void  generateJaxbClass(String wsdlUrl, String rootPath, String genDirName) {
		try {
			String generatedDir = rootPath + File.separator + genDirName;
			Util.createDirectory(generatedDir);
			Driver.run(createJxcOptions(wsdlUrl,generatedDir), new PEtALSXjcListener());	
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	
	/**
	 * Compile the generated JAXB classes 
	 * @param baseDirPath : The base directory of the compilation
	 * @param srcDirName : The source directory name
	 * @param classPath : List of class path to compile the source file
	 */
	public static void compileJaxbClass(String baseDirPath, String srcDirName, List<String> classPath) {
		
    	try {
    		DataBindingCompiler.compile(baseDirPath, srcDirName, classPath);
		} catch (Exception e) {
			e.printStackTrace();
		}
    	
		
	}
	
	
	/**
	 * Create option for JAXB to generate databinding classes 
	 * @param wsdlUrl : The WSDL URL
	 * @param generatedDir : The name of the directory where the sources will be generated
	 * @return a list of options
	 */
	private static String[] createJxcOptions(String wsdlUrl, String generatedDir) {
		
		String[] options = new String[4];
		options[0] = "-wsdl";
		options[1] = wsdlUrl;
		options[2] = "-d";
		options[3] = generatedDir;
		
		return options;
	}
	
	

}
